"""CLI command for installing NiftyReg binaries."""

from pathlib import Path
from typing import Annotated, Optional

import typer
from loguru import logger

from niftyregw.commands import setup_logger
from niftyregw.enums import LogLevel
from niftyregw.install import _DEFAULT_OUTPUT_DIR, download_niftyreg, get_platform


def install(
    output_dir: Annotated[
        Optional[Path],
        typer.Option(
            "--output-dir",
            "-o",
            help="Directory to install binaries into. Default: ~/.local/bin.",
        ),
    ] = None,
    show_platform: Annotated[
        bool,
        typer.Option(
            "--platform",
            help="Show the detected platform and exit without installing.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Download and install NiftyReg binaries."""
    setup_logger(log_level)
    install_logger = logger.bind(executable="niftyregw")

    if show_platform:
        platform_name = get_platform()
        install_logger.info(f"Platform: {platform_name}")
        return

    out_dir = output_dir if output_dir is not None else _DEFAULT_OUTPUT_DIR
    install_logger.info(f"Downloading NiftyReg binaries to {out_dir}...")
    installed = download_niftyreg(out_dir)
    for path in installed:
        install_logger.info(f"  Installed {path.name} → {path}")
    install_logger.info(f"Done! {len(installed)} binaries installed.")



"""CLI command for reg_aladin."""

from pathlib import Path
from typing import Annotated, Optional

import typer

from niftyregw.commands import make_help_callback, make_version_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import reg_aladin as _reg_aladin

_help_callback = make_help_callback("reg_aladin")
_version_callback = make_version_callback("reg_aladin")


def aladin(
    reference: Annotated[
        Path,
        typer.Option(
            "--reference",
            "-r",
            help="Reference image filename (also called Target or Fixed).",
        ),
    ],
    floating: Annotated[
        Path,
        typer.Option(
            "--floating",
            "-f",
            help="Floating image filename (also called Source or Moving).",
        ),
    ],
    # Output options
    output_affine: Annotated[
        Optional[Path],
        typer.Option(
            "--output-affine",
            "-a",
            help="Output affine transformation filename. [outputAffine.txt]",
        ),
    ] = None,
    output_result: Annotated[
        Optional[Path],
        typer.Option(
            "--output-result",
            "-o",
            help="Resampled image filename. [outputResult.nii.gz]",
        ),
    ] = None,
    # Input options
    input_affine: Annotated[
        Optional[Path],
        typer.Option(
            help="Input affine transformation filename. (Affine*Reference=Floating)",
        ),
    ] = None,
    reference_mask: Annotated[
        Optional[Path],
        typer.Option(
            help="Mask image in the reference space.",
        ),
    ] = None,
    floating_mask: Annotated[
        Optional[Path],
        typer.Option(
            help="Mask image in the floating space. (Only used when symmetric turned on)"
        ),
    ] = None,
    # Algorithm options
    no_symmetric: Annotated[
        bool, typer.Option(help="Disable the symmetric version of the algorithm.")
    ] = False,
    rigid_only: Annotated[
        bool,
        typer.Option(
            help="Perform a rigid registration only. (Rigid+affine by default)"
        ),
    ] = False,
    affine_direct: Annotated[
        bool, typer.Option(help="Directly optimize 12 DoF affine.")
    ] = False,
    max_iterations: Annotated[
        Optional[int],
        typer.Option(
            help="Max iterations of the trimmed least square approach per level. [5]"
        ),
    ] = None,
    num_levels: Annotated[
        Optional[int],
        typer.Option(help="Number of levels for the coarse-to-fine pyramids. [3]"),
    ] = None,
    num_levels_to_perform: Annotated[
        Optional[int],
        typer.Option(help="Number of levels to run the registration. [ln]"),
    ] = None,
    smooth_reference: Annotated[
        Optional[float],
        typer.Option(
            help="Gaussian smoothing std dev (mm) for the reference image. [0]"
        ),
    ] = None,
    smooth_floating: Annotated[
        Optional[float],
        typer.Option(
            help="Gaussian smoothing std dev (mm) for the floating image. [0]"
        ),
    ] = None,
    reference_lower_threshold: Annotated[
        Optional[float],
        typer.Option(help="Lower threshold for the reference image. [0]"),
    ] = None,
    reference_upper_threshold: Annotated[
        Optional[float],
        typer.Option(help="Upper threshold for the reference image. [0]"),
    ] = None,
    floating_lower_threshold: Annotated[
        Optional[float],
        typer.Option(help="Lower threshold for the floating image. [0]"),
    ] = None,
    floating_upper_threshold: Annotated[
        Optional[float],
        typer.Option(help="Upper threshold for the floating image. [0]"),
    ] = None,
    padding: Annotated[
        Optional[float], typer.Option(help="Padding value. [nan]")
    ] = None,
    # Initialisation options
    use_nifti_origin: Annotated[
        bool,
        typer.Option(
            help="Use the nifti header origin to initialise the transformation."
        ),
    ] = False,
    use_masks_centre_of_mass: Annotated[
        bool,
        typer.Option(
            help="Use the input masks centre of mass to initialise the transformation."
        ),
    ] = False,
    use_images_centre_of_mass: Annotated[
        bool,
        typer.Option(
            help="Use the input images centre of mass to initialise the transformation."
        ),
    ] = False,
    # Other options
    interpolation: Annotated[
        Optional[int],
        typer.Option(help="Interpolation order to warp the floating image."),
    ] = None,
    isotropic: Annotated[
        bool,
        typer.Option(help="Make floating and reference images isotropic if required."),
    ] = False,
    percent_blocks_to_use: Annotated[
        Optional[int],
        typer.Option(
            help="Percentage of blocks to use in the optimisation scheme. [50]"
        ),
    ] = None,
    percent_inliers: Annotated[
        Optional[int],
        typer.Option(help="Percentage of blocks to consider as inlier. [50]"),
    ] = None,
    block_step_size_2: Annotated[
        bool,
        typer.Option(
            help="Use a block step size of 2 (instead of 1) for faster, less accurate registration."
        ),
    ] = False,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of threads to use with OpenMP.")
    ] = None,
    verbose_off: Annotated[bool, typer.Option(help="Turn verbose off.")] = False,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            is_eager=True,
            callback=_version_callback,
            help="Print reg_aladin version and exit.",
        ),
    ] = False,
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_aladin help and exit.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Block-matching global (affine/rigid) registration."""
    setup_logger(log_level)

    _reg_aladin(
        reference,
        floating,
        output_affine=output_affine,
        output_result=output_result,
        input_affine=input_affine,
        reference_mask=reference_mask,
        floating_mask=floating_mask,
        no_symmetric=no_symmetric,
        rigid_only=rigid_only,
        affine_direct=affine_direct,
        max_iterations=max_iterations,
        num_levels=num_levels,
        num_levels_to_perform=num_levels_to_perform,
        smooth_reference=smooth_reference,
        smooth_floating=smooth_floating,
        reference_lower_threshold=reference_lower_threshold,
        reference_upper_threshold=reference_upper_threshold,
        floating_lower_threshold=floating_lower_threshold,
        floating_upper_threshold=floating_upper_threshold,
        padding=padding,
        use_nifti_origin=use_nifti_origin,
        use_masks_centre_of_mass=use_masks_centre_of_mass,
        use_images_centre_of_mass=use_images_centre_of_mass,
        interpolation=interpolation,
        isotropic=isotropic,
        percent_blocks_to_use=percent_blocks_to_use,
        percent_inliers=percent_inliers,
        block_step_size_2=block_step_size_2,
        omp_threads=omp_threads,
        verbose_off=verbose_off,
    )



"""CLI command for reg_average."""

from pathlib import Path
from typing import Annotated, List

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Average images or transformations.",
)

_help_callback = make_help_callback("reg_average")


@app.callback()
def _main(
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_average help and exit.",
        ),
    ] = False,
) -> None:
    pass


@app.command("avg")
def avg(
    output: Annotated[Path, typer.Option("--output", "-o", help="Output filename.")],
    inputs: Annotated[
        List[Path], typer.Argument(help="Input images or affine matrices to average.")
    ],
    nearest_neighbour: Annotated[
        bool, typer.Option("--nn", help="Use nearest neighbour interpolation.")
    ] = False,
    linear: Annotated[
        bool, typer.Option("--lin", help="Use linear interpolation.")
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Average images or affine matrices."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_average")
    args: list[str] = [str(output), "-avg"] + [str(p) for p in inputs]
    if nearest_neighbour:
        args.append("--NN")
    if linear:
        args.append("--LIN")
    run("reg_average", *args, tool_logger=tool_logger)


@app.command("avg-lts")
def avg_lts(
    output: Annotated[Path, typer.Option("--output", "-o", help="Output filename.")],
    inputs: Annotated[
        List[Path],
        typer.Argument(help="Input affine matrices (half treated as outliers)."),
    ],
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Robust average of affine matrices (LTS, half are outliers)."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_average")
    args: list[str] = [str(output), "-avg_lts"] + [str(p) for p in inputs]
    run("reg_average", *args, tool_logger=tool_logger)


@app.command("avg-tran")
def avg_tran(
    output: Annotated[Path, typer.Option("--output", "-o", help="Output filename.")],
    reference: Annotated[
        Path, typer.Option("--reference", "-r", help="Reference image.")
    ],
    inputs: Annotated[
        List[Path],
        typer.Argument(
            help="Alternating transformation and floating image files: trans1 flo1 trans2 flo2 ..."
        ),
    ],
    nearest_neighbour: Annotated[
        bool, typer.Option("--nn", help="Use nearest neighbour interpolation.")
    ] = False,
    linear: Annotated[
        bool, typer.Option("--lin", help="Use linear interpolation.")
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Resample all images into reference space and average (cubic spline)."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_average")
    args: list[str] = [str(output), "-avg_tran", str(reference)] + [
        str(p) for p in inputs
    ]
    if nearest_neighbour:
        args.append("--NN")
    if linear:
        args.append("--LIN")
    run("reg_average", *args, tool_logger=tool_logger)


@app.command("demean")
def demean(
    output: Annotated[Path, typer.Option("--output", "-o", help="Output filename.")],
    reference: Annotated[
        Path, typer.Option("--reference", "-r", help="Reference image.")
    ],
    inputs: Annotated[
        List[Path],
        typer.Argument(
            help="Alternating transformation and floating image files: trans1 flo1 trans2 flo2 ..."
        ),
    ],
    nearest_neighbour: Annotated[
        bool, typer.Option("--nn", help="Use nearest neighbour interpolation.")
    ] = False,
    linear: Annotated[
        bool, typer.Option("--lin", help="Use linear interpolation.")
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Demean transformations (enforce mean identity)."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_average")
    args: list[str] = [str(output), "-demean", str(reference)] + [
        str(p) for p in inputs
    ]
    if nearest_neighbour:
        args.append("--NN")
    if linear:
        args.append("--LIN")
    run("reg_average", *args, tool_logger=tool_logger)


@app.command("demean-noaff")
def demean_noaff(
    output: Annotated[Path, typer.Option("--output", "-o", help="Output filename.")],
    reference: Annotated[
        Path, typer.Option("--reference", "-r", help="Reference image.")
    ],
    inputs: Annotated[
        List[Path],
        typer.Argument(
            help="Triples of affine, non-rigid transformation, and floating image: aff1 nr1 flo1 ..."
        ),
    ],
    nearest_neighbour: Annotated[
        bool, typer.Option("--nn", help="Use nearest neighbour interpolation.")
    ] = False,
    linear: Annotated[
        bool, typer.Option("--lin", help="Use linear interpolation.")
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Demean with explicit affine removal from non-rigid transformations."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_average")
    args: list[str] = [str(output), "-demean_noaff", str(reference)] + [
        str(p) for p in inputs
    ]
    if nearest_neighbour:
        args.append("--NN")
    if linear:
        args.append("--LIN")
    run("reg_average", *args, tool_logger=tool_logger)


@app.command("cmd-file")
def cmd_file(
    output: Annotated[Path, typer.Option("--output", "-o", help="Output filename.")],
    command_file: Annotated[
        Path,
        typer.Option(
            "--command-file", "-c", help="Text file containing the full command."
        ),
    ],
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Run reg_average from a command file."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_average")
    run(
        "reg_average",
        str(output),
        "--cmd_file",
        str(command_file),
        tool_logger=tool_logger,
    )



"""Python wrapper for NiftyReg binaries."""

from __future__ import annotations

from pathlib import Path
from subprocess import PIPE, Popen
from threading import Thread
from typing import TextIO

import loguru
from loguru import logger

from .install import find as _find

# Matrix formatting constants
_MATRIX_COLUMN_COUNT = 4


def _format_matrix_line(line: str) -> str:
    """Format a line that looks like a transformation matrix row.

    Detects lines with exactly 4 numbers separated by whitespace and formats them
    with consistent spacing and precision.

    Args:
        line: The line to potentially format.

    Returns:
        The formatted line if it's a matrix row, otherwise the original line.
    """
    parts = line.split()

    # Check if this line has exactly 4 parts that are all numbers
    if len(parts) != _MATRIX_COLUMN_COUNT:
        return line

    try:
        numbers = [float(part) for part in parts]
    except ValueError:
        # Not all parts are numbers, return original
        return line

    # Format all numbers with 3 decimal places and minimum width of 7 characters
    # Width of 7 accommodates: 3 digits + 1 decimal point + 3 decimals
    # Numbers larger than 999.999 will expand beyond the minimum width
    formatted_parts = [f"{num:7.3f}" for num in numbers]

    # Add leading space and join with two spaces between columns
    return " " + "  ".join(formatted_parts)


def _get_path(tool: str) -> Path:
    path = _find(tool)
    if path is None:
        msg = f"{tool} not found. Please install NiftyReg first."
        raise FileNotFoundError(msg)
    return path


def _read_stream(
    stream: TextIO,
    is_stderr: bool,
    tool_logger: loguru.Logger | None,
) -> None:
    """Read lines from a stream and log them appropriately.

    Args:
        stream: The stream to read from.
        is_stderr: Whether this is the stderr stream.
        tool_logger: Optional loguru logger for structured output.
    """
    for line in stream:
        line = line.rstrip("\n")
        # Format matrix lines for better readability
        line = _format_matrix_line(line)
        if tool_logger is None:
            logger.info(line)
            continue

        # Determine log level and strip NiftyReg prefix
        message = line
        if is_stderr:
            if line.startswith("[NiftyReg WARNING]"):
                log = tool_logger.warning
                message = line[len("[NiftyReg WARNING]") :].lstrip()
            elif line.startswith("[NiftyReg ERROR]"):
                log = tool_logger.error
                message = line[len("[NiftyReg ERROR]") :].lstrip()
            else:
                log = tool_logger.info
        else:
            log = tool_logger.info

        # Strip [NiftyReg INFO] prefix from stdout and remaining stderr lines
        if message.startswith("[NiftyReg INFO]"):
            message = message[len("[NiftyReg INFO]") :].lstrip()

        log(message)


def run(tool: str, *args: str, tool_logger: loguru.Logger | None = None) -> None:
    """Run any NiftyReg binary with raw CLI arguments.

    Args:
        tool: Binary name (e.g. ``"reg_aladin"``).
        *args: Raw CLI arguments.
        tool_logger: Optional loguru logger for structured output.
    """
    tool_path = _get_path(tool)
    args_list = [arg.strip("\\\n") for arg in args]
    args_list = [arg for arg in args_list if arg]

    cmd = [str(tool_path), *args_list]
    with Popen(cmd, stdout=PIPE, stderr=PIPE, text=True, bufsize=1) as p:
        assert p.stdout is not None
        assert p.stderr is not None

        # Read both streams concurrently using threads
        stderr_thread = Thread(
            target=_read_stream,
            args=(p.stderr, True, tool_logger),
            name="stderr-reader",
        )
        stdout_thread = Thread(
            target=_read_stream,
            args=(p.stdout, False, tool_logger),
            name="stdout-reader",
        )

        stderr_thread.start()
        stdout_thread.start()

        # Wait for both threads to complete
        stderr_thread.join()
        stdout_thread.join()

        # Wait for process to complete
        p.wait()


def reg_aladin(
    reference: Path,
    floating: Path,
    *,
    output_affine: Path | None = None,
    output_result: Path | None = None,
    input_affine: Path | None = None,
    reference_mask: Path | None = None,
    floating_mask: Path | None = None,
    no_symmetric: bool = False,
    rigid_only: bool = False,
    affine_direct: bool = False,
    max_iterations: int | None = None,
    num_levels: int | None = None,
    num_levels_to_perform: int | None = None,
    smooth_reference: float | None = None,
    smooth_floating: float | None = None,
    reference_lower_threshold: float | None = None,
    reference_upper_threshold: float | None = None,
    floating_lower_threshold: float | None = None,
    floating_upper_threshold: float | None = None,
    padding: float | None = None,
    use_nifti_origin: bool = False,
    use_masks_centre_of_mass: bool = False,
    use_images_centre_of_mass: bool = False,
    interpolation: int | None = None,
    isotropic: bool = False,
    percent_blocks_to_use: int | None = None,
    percent_inliers: int | None = None,
    block_step_size_2: bool = False,
    omp_threads: int | None = None,
    verbose_off: bool = False,
) -> None:
    """Run reg_aladin with structured arguments.

    Args:
        reference: Reference image path (also called Target or Fixed).
        floating: Floating image path (also called Source or Moving).
        output_affine: Output affine transformation filename.
        output_result: Resampled image filename.
        input_affine: Input affine transformation filename.
        reference_mask: Mask image in the reference space.
        floating_mask: Mask image in the floating space.
        no_symmetric: Disable the symmetric version of the algorithm.
        rigid_only: Perform a rigid registration only.
        affine_direct: Directly optimize 12 DoF affine.
        max_iterations: Max iterations per level.
        num_levels: Number of levels for pyramids.
        num_levels_to_perform: Number of levels to run.
        smooth_reference: Gaussian smoothing std dev (mm) for reference.
        smooth_floating: Gaussian smoothing std dev (mm) for floating.
        reference_lower_threshold: Lower threshold for reference image.
        reference_upper_threshold: Upper threshold for reference image.
        floating_lower_threshold: Lower threshold for floating image.
        floating_upper_threshold: Upper threshold for floating image.
        padding: Padding value.
        use_nifti_origin: Use nifti header origin for initialisation.
        use_masks_centre_of_mass: Use masks centre of mass for initialisation.
        use_images_centre_of_mass: Use images centre of mass for initialisation.
        interpolation: Interpolation order.
        isotropic: Make images isotropic if required.
        percent_blocks_to_use: Percentage of blocks in optimisation.
        percent_inliers: Percentage of inlier blocks.
        block_step_size_2: Use block step size of 2 for faster registration.
        omp_threads: Number of OpenMP threads.
        verbose_off: Turn verbose off.
    """
    # Check if the default output file exists before registration
    default_output = Path("outputAffine.txt")
    existed_before = default_output.exists()

    # Determine if user requested the default output path
    user_requested_default = False
    if output_affine is not None:
        try:
            # Resolve both paths to absolute paths for comparison
            user_path = Path(output_affine).resolve()
            default_path = default_output.resolve()
            user_requested_default = user_path == default_path
        except (OSError, RuntimeError):
            # If path resolution fails, treat as different paths
            pass

    command_lines: list[str] = [
        f"  -ref {reference} \\",
        f"  -flo {floating} \\",
    ]

    if output_affine is not None:
        command_lines.append(f"  -aff {output_affine} \\")
    if output_result is not None:
        command_lines.append(f"  -res {output_result} \\")
    if input_affine is not None:
        command_lines.append(f"  -inaff {input_affine} \\")
    if reference_mask is not None:
        command_lines.append(f"  -rmask {reference_mask} \\")
    if floating_mask is not None:
        command_lines.append(f"  -fmask {floating_mask} \\")
    if no_symmetric:
        command_lines.append("  -noSym \\")
    if rigid_only:
        command_lines.append("  -rigOnly \\")
    if affine_direct:
        command_lines.append("  -affDirect \\")
    if max_iterations is not None:
        command_lines.append(f"  -maxit {max_iterations} \\")
    if num_levels is not None:
        command_lines.append(f"  -ln {num_levels} \\")
    if num_levels_to_perform is not None:
        command_lines.append(f"  -lp {num_levels_to_perform} \\")
    if smooth_reference is not None:
        command_lines.append(f"  -smooR {smooth_reference} \\")
    if smooth_floating is not None:
        command_lines.append(f"  -smooF {smooth_floating} \\")
    if reference_lower_threshold is not None:
        command_lines.append(f"  -refLowThr {reference_lower_threshold} \\")
    if reference_upper_threshold is not None:
        command_lines.append(f"  -refUpThr {reference_upper_threshold} \\")
    if floating_lower_threshold is not None:
        command_lines.append(f"  -floLowThr {floating_lower_threshold} \\")
    if floating_upper_threshold is not None:
        command_lines.append(f"  -floUpThr {floating_upper_threshold} \\")
    if padding is not None:
        command_lines.append(f"  -pad {padding} \\")
    if use_nifti_origin:
        command_lines.append("  -nac \\")
    if use_masks_centre_of_mass:
        command_lines.append("  -comm \\")
    if use_images_centre_of_mass:
        command_lines.append("  -comi \\")
    if interpolation is not None:
        command_lines.append(f"  -interp {interpolation} \\")
    if isotropic:
        command_lines.append("  -iso \\")
    if percent_blocks_to_use is not None:
        command_lines.append(f"  -pv {percent_blocks_to_use} \\")
    if percent_inliers is not None:
        command_lines.append(f"  -pi {percent_inliers} \\")
    if block_step_size_2:
        command_lines.append("  -speeeeed \\")
    if omp_threads is not None:
        command_lines.append(f"  -omp {omp_threads} \\")
    if verbose_off:
        command_lines.append("  -voff \\")

    _run_with_logging("reg_aladin", *command_lines)

    # Clean up the default output file if it was created and not requested
    if not existed_before and default_output.exists() and not user_requested_default:
        try:
            default_output.unlink()
            logger.bind(executable="niftyregw").debug(
                f"Cleaned up default output file: {default_output}"
            )
        except OSError as e:
            logger.bind(executable="niftyregw").warning(
                f"Failed to clean up {default_output}: {e}"
            )


def _run_with_logging(tool: str, *lines: str) -> None:
    tool_path = _get_path(tool)
    loggerw = logger.bind(executable="niftyregw")
    loggerx = logger.bind(executable=tool)

    loggerw.debug("The following command will be run:")
    lines_str = "\n".join(lines).strip(" \\")
    loggerw.debug(f"{tool_path} \\\n  {lines_str}")
    args = []
    for line in lines:
        args.extend(line.strip(" \\").split())

    run(tool, *args, tool_logger=loggerx)



"""Typer CLI for all NiftyReg binaries."""

import typer

from niftyregw.commands.aladin import aladin
from niftyregw.commands.average import app as average_app
from niftyregw.commands.f3d import f3d
from niftyregw.commands.install import install
from niftyregw.commands.jacobian import jacobian
from niftyregw.commands.measure import measure
from niftyregw.commands.resample import resample
from niftyregw.commands.tools import tools
from niftyregw.commands.transform import app as transform_app

app = typer.Typer(add_completion=False, no_args_is_help=True)

app.command(
    "install",
    help="Download and install NiftyReg binaries.",
)(install)

app.command(
    "aladin",
    help="Block-matching global (affine/rigid) registration.",
    no_args_is_help=True,
)(aladin)
app.command(
    "f3d",
    help="Fast Free-Form Deformation (F3D) non-rigid registration.",
    no_args_is_help=True,
)(f3d)
app.command(
    "measure", help="Compute similarity measures between images.", no_args_is_help=True
)(measure)
app.command(
    "jacobian",
    help="Compute Jacobian-based maps from transformations.",
    no_args_is_help=True,
)(jacobian)
app.command(
    "resample",
    help="Resample an image with a given transformation.",
    no_args_is_help=True,
)(resample)
app.command("tools", help="Image manipulation tools.", no_args_is_help=True)(tools)
app.add_typer(average_app, name="average")
app.add_typer(transform_app, name="transform")


if __name__ == "__main__":
    app()



"""CLI command for reg_jacobian."""

from pathlib import Path
from typing import Annotated, Optional

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, make_version_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

_help_callback = make_help_callback("reg_jacobian")
_version_callback = make_version_callback("reg_jacobian")


def jacobian(
    transformation: Annotated[
        Path,
        typer.Option("--transformation", "-t", help="Transformation file (mandatory)."),
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    jacobian_determinant: Annotated[
        Optional[Path], typer.Option(help="Output Jacobian determinant map.")
    ] = None,
    jacobian_matrix: Annotated[
        Optional[Path], typer.Option(help="Output Jacobian matrix map (5D nifti).")
    ] = None,
    jacobian_log_determinant: Annotated[
        Optional[Path], typer.Option(help="Output log of Jacobian determinant map.")
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of threads to use with OpenMP.")
    ] = None,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            is_eager=True,
            callback=_version_callback,
            help="Print version and exit.",
        ),
    ] = False,
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_jacobian help and exit.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Compute Jacobian-based maps from a transformation."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_jacobian")

    args: list[str] = ["-trans", str(transformation)]
    if reference is not None:
        args.extend(["-ref", str(reference)])
    if jacobian_determinant is not None:
        args.extend(["-jac", str(jacobian_determinant)])
    if jacobian_matrix is not None:
        args.extend(["-jacM", str(jacobian_matrix)])
    if jacobian_log_determinant is not None:
        args.extend(["-jacL", str(jacobian_log_determinant)])
    if omp_threads is not None:
        args.extend(["-omp", str(omp_threads)])

    run("reg_jacobian", *args, tool_logger=tool_logger)



from importlib.metadata import version

from .install import download_niftyreg, get_platform
from .wrapper import reg_aladin, run

__all__ = [
    "download_niftyreg",
    "get_platform",
    "reg_aladin",
    "run",
]

assert __package__ is not None
__version__ = version(__package__)



"""CLI command for reg_measure."""

from pathlib import Path
from typing import Annotated, Optional

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, make_version_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

_help_callback = make_help_callback("reg_measure")
_version_callback = make_version_callback("reg_measure")


def measure(
    reference: Annotated[
        Path, typer.Option("--reference", "-r", help="Reference image filename.")
    ],
    floating: Annotated[
        Path, typer.Option("--floating", "-f", help="Floating image filename.")
    ],
    ncc: Annotated[
        bool, typer.Option(help="Compute NCC (Normalized Cross-Correlation).")
    ] = False,
    lncc: Annotated[bool, typer.Option(help="Compute LNCC (Local NCC).")] = False,
    nmi: Annotated[
        bool, typer.Option(help="Compute NMI (Normalized Mutual Information, 64 bins).")
    ] = False,
    ssd: Annotated[
        bool, typer.Option(help="Compute SSD (Sum of Squared Differences).")
    ] = False,
    output: Annotated[
        Optional[Path],
        typer.Option("--output", "-o", help="Output text file. [stdout]"),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of threads to use with OpenMP.")
    ] = None,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            is_eager=True,
            callback=_version_callback,
            help="Print version and exit.",
        ),
    ] = False,
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_measure help and exit.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Compute similarity measures between two images."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_measure")

    args: list[str] = ["-ref", str(reference), "-flo", str(floating)]
    if ncc:
        args.append("-ncc")
    if lncc:
        args.append("-lncc")
    if nmi:
        args.append("-nmi")
    if ssd:
        args.append("-ssd")
    if output is not None:
        args.extend(["-out", str(output)])
    if omp_threads is not None:
        args.extend(["-omp", str(omp_threads)])

    run("reg_measure", *args, tool_logger=tool_logger)



"""CLI command for reg_transform."""

from pathlib import Path
from typing import Annotated, Optional, Tuple

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
    help="Manipulate and compose transformations.",
)

_help_callback = make_help_callback("reg_transform")


@app.callback()
def _main(
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_transform help and exit.",
        ),
    ] = False,
) -> None:
    pass


# ---------------------------------------------------------------------------
# Shared option patterns
# ---------------------------------------------------------------------------


def _run_transform(
    args: list[str], log_level: LogLevel, omp_threads: Optional[int] = None
) -> None:
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_transform")
    if omp_threads is not None:
        args.extend(["-omp", str(omp_threads)])
    run("reg_transform", *args, tool_logger=tool_logger)


# ---------------------------------------------------------------------------
# Sub-commands
# ---------------------------------------------------------------------------


@app.command("deformation")
def deformation(
    input_transformation: Annotated[
        Path, typer.Option("--input", "-i", help="Input transformation file.")
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output deformation field file.")
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Compute a deformation field from a transformation."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    args.extend(["-def", str(input_transformation), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("displacement")
def displacement(
    input_transformation: Annotated[
        Path, typer.Option("--input", "-i", help="Input transformation file.")
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output displacement field file.")
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Compute a displacement field from a transformation."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    args.extend(["-disp", str(input_transformation), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("flow")
def flow(
    input_transformation: Annotated[
        Path, typer.Option("--input", "-i", help="Input spline parametrised SVF.")
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output flow field file.")
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Compute a flow field from a spline parametrised SVF."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    args.extend(["-flow", str(input_transformation), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("compose")
def compose(
    input1: Annotated[
        Path, typer.Option("--input1", "-i", help="First transformation file.")
    ],
    input2: Annotated[
        Path, typer.Option("--input2", "-j", help="Second transformation file.")
    ],
    output: Annotated[
        Path,
        typer.Option(
            "--output", "-o", help="Output deformation field: T3(x) = T2(T1(x))."
        ),
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference", "-r", help="Reference image for input1 (if spline)."
        ),
    ] = None,
    reference2: Annotated[
        Optional[Path], typer.Option(help="Reference image for input2 (if spline).")
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Compose two transformations into a deformation field."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    if reference2 is not None:
        args.extend(["-ref2", str(reference2)])
    args.extend(["-comp", str(input1), str(input2), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("landmarks")
def landmarks(
    transformation: Annotated[
        Path, typer.Option("--transformation", "-t", help="Transformation file.")
    ],
    input_landmarks: Annotated[
        Path, typer.Option("--input", "-i", help="Input landmark file (mm positions).")
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output landmark file.")
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Apply a transformation to a set of landmarks."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    args.extend(["-land", str(transformation), str(input_landmarks), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("update-sform")
def update_sform(
    input_image: Annotated[
        Path, typer.Option("--input", "-i", help="Image to be updated.")
    ],
    affine: Annotated[
        Path,
        typer.Option(
            "--affine", "-a", help="Affine transformation (Affine*Reference=Floating)."
        ),
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output updated image.")
    ],
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Update the sform of an image using an affine transformation."""
    args: list[str] = ["-updSform", str(input_image), str(affine), str(output)]
    _run_transform(args, log_level, omp_threads)


@app.command("invert-affine")
def invert_affine(
    input_affine: Annotated[
        Path, typer.Option("--input", "-i", help="Input affine transformation file.")
    ],
    output: Annotated[
        Path,
        typer.Option(
            "--output", "-o", help="Output inverted affine transformation file."
        ),
    ],
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Invert an affine matrix."""
    args: list[str] = ["-invAff", str(input_affine), str(output)]
    _run_transform(args, log_level, omp_threads)


@app.command("invert-nonrigid")
def invert_nonrigid(
    input_transformation: Annotated[
        Path, typer.Option("--input", "-i", help="Input transformation file.")
    ],
    floating: Annotated[
        Path,
        typer.Option(
            "--floating",
            "-f",
            help="Floating image where inverted transformation is defined.",
        ),
    ],
    output: Annotated[
        Path,
        typer.Option("--output", "-o", help="Output inverted transformation file."),
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Invert a non-rigid transformation (saved as deformation field)."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    args.extend(["-invNrr", str(input_transformation), str(floating), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("half")
def half(
    input_transformation: Annotated[
        Path, typer.Option("--input", "-i", help="Input transformation file.")
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output halved transformation file.")
    ],
    reference: Annotated[
        Optional[Path],
        typer.Option(
            "--reference",
            "-r",
            help="Reference image (required for spline parametrisation).",
        ),
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Halve a transformation (same type preserved)."""
    args: list[str] = []
    if reference is not None:
        args.extend(["-ref", str(reference)])
    args.extend(["-half", str(input_transformation), str(output)])
    _run_transform(args, log_level, omp_threads)


@app.command("make-affine")
def make_affine(
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output affine transformation file.")
    ],
    rotation: Annotated[
        Tuple[float, float, float], typer.Option(help="Rotation angles (rx ry rz).")
    ],
    translation: Annotated[
        Tuple[float, float, float], typer.Option(help="Translation (tx ty tz).")
    ],
    scaling: Annotated[
        Tuple[float, float, float], typer.Option(help="Scaling factors (sx sy sz).")
    ],
    shearing: Annotated[
        Tuple[float, float, float], typer.Option(help="Shearing (shx shy shz).")
    ],
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Create an affine transformation matrix from parameters."""
    params = [
        str(rotation[0]),
        str(rotation[1]),
        str(rotation[2]),
        str(translation[0]),
        str(translation[1]),
        str(translation[2]),
        str(scaling[0]),
        str(scaling[1]),
        str(scaling[2]),
        str(shearing[0]),
        str(shearing[1]),
        str(shearing[2]),
    ]
    args: list[str] = ["-makeAff", *params, str(output)]
    _run_transform(args, log_level, omp_threads)


@app.command("affine-to-rigid")
def affine_to_rigid(
    input_affine: Annotated[
        Path, typer.Option("--input", "-i", help="Input affine transformation file.")
    ],
    output: Annotated[
        Path, typer.Option("--output", "-o", help="Output rigid transformation file.")
    ],
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Extract the rigid component from an affine transformation."""
    args: list[str] = ["-aff2rig", str(input_affine), str(output)]
    _run_transform(args, log_level, omp_threads)


@app.command("flirt-to-niftyreg")
def flirt_to_niftyreg(
    input_flirt: Annotated[
        Path,
        typer.Option("--input", "-i", help="Input FLIRT (FSL) affine transformation."),
    ],
    reference: Annotated[
        Path,
        typer.Option("--reference", "-r", help="Reference image used in FLIRT (-ref)."),
    ],
    floating: Annotated[
        Path,
        typer.Option("--floating", "-f", help="Floating image used in FLIRT (-in)."),
    ],
    output: Annotated[
        Path,
        typer.Option("--output", "-o", help="Output NiftyReg affine transformation."),
    ],
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of OpenMP threads.")
    ] = None,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Convert a FLIRT (FSL) affine to a NiftyReg affine."""
    args: list[str] = [
        "-flirtAff2NR",
        str(input_flirt),
        str(reference),
        str(floating),
        str(output),
    ]
    _run_transform(args, log_level, omp_threads)



"""Shared helpers for CLI commands."""

import sys

import typer
from loguru import logger

from niftyregw.enums import LogLevel
from niftyregw.wrapper import run


def setup_logger(log_level: LogLevel) -> None:
    """Configure loguru for CLI output."""
    logger.remove()
    logger.add(
        sys.stderr,
        format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{extra[executable]}</level> | <level>{level: <8}</level> | <level>{message}</level>",
        level=log_level.value,
        colorize=True,
    )


def make_help_callback(tool: str):
    """Create an eager Typer callback that shows the original binary help."""

    def _callback(value: bool) -> None:
        if value:
            run(tool, "-h")
            raise typer.Exit()

    return _callback


def make_version_callback(tool: str):
    """Create an eager Typer callback that shows the original binary version."""

    def _callback(value: bool) -> None:
        if value:
            run(tool, "--version")
            raise typer.Exit()

    return _callback



"""CLI command for reg_f3d."""

from pathlib import Path
from typing import Annotated, Optional

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, make_version_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

_help_callback = make_help_callback("reg_f3d")
_version_callback = make_version_callback("reg_f3d")


def f3d(
    reference: Annotated[
        Path, typer.Option("--reference", "-r", help="Reference image filename.")
    ],
    floating: Annotated[
        Path, typer.Option("--floating", "-f", help="Floating image filename.")
    ],
    # Initial transformation (one of)
    input_affine: Annotated[
        Optional[Path],
        typer.Option(
            "--input-affine",
            "-a",
            help="Input affine transformation (Affine*Reference=Floating).",
        ),
    ] = None,
    input_cpp: Annotated[
        Optional[Path],
        typer.Option(help="Input control point grid (defines coarse spacing)."),
    ] = None,
    # Output
    output_cpp: Annotated[
        Optional[Path], typer.Option(help="Output control point grid. [outputCPP.nii]")
    ] = None,
    output_result: Annotated[
        Optional[Path],
        typer.Option(
            "--output-result", "-o", help="Resampled image. [outputResult.nii]"
        ),
    ] = None,
    # Input image options
    reference_mask: Annotated[
        Optional[Path],
        typer.Option(
            "--reference-mask", "-m", help="Mask image in the reference space."
        ),
    ] = None,
    smooth_reference: Annotated[
        Optional[float], typer.Option(help="Smooth reference image (sigma in mm). [0]")
    ] = None,
    smooth_floating: Annotated[
        Optional[float], typer.Option(help="Smooth floating image (sigma in mm). [0]")
    ] = None,
    reference_lower_threshold: Annotated[
        Optional[float],
        typer.Option(help="Lower threshold for reference image intensities."),
    ] = None,
    reference_upper_threshold: Annotated[
        Optional[float],
        typer.Option(help="Upper threshold for reference image intensities."),
    ] = None,
    floating_lower_threshold: Annotated[
        Optional[float],
        typer.Option(help="Lower threshold for floating image intensities."),
    ] = None,
    floating_upper_threshold: Annotated[
        Optional[float],
        typer.Option(help="Upper threshold for floating image intensities."),
    ] = None,
    # Spline options
    spacing_x: Annotated[
        Optional[float],
        typer.Option(
            help="Final grid spacing along x in mm (voxels if negative). [5 vox]"
        ),
    ] = None,
    spacing_y: Annotated[
        Optional[float],
        typer.Option(
            help="Final grid spacing along y in mm (voxels if negative). [sx]"
        ),
    ] = None,
    spacing_z: Annotated[
        Optional[float],
        typer.Option(
            help="Final grid spacing along z in mm (voxels if negative). [sx]"
        ),
    ] = None,
    # Regularisation
    bending_energy: Annotated[
        Optional[float], typer.Option(help="Bending energy penalty weight. [0.001]")
    ] = None,
    linear_energy: Annotated[
        Optional[float],
        typer.Option(help="Linear energy (first order) penalty weight. [0.01]"),
    ] = None,
    jacobian_log_weight: Annotated[
        Optional[float],
        typer.Option(help="Log of Jacobian determinant penalty weight. [0.0]"),
    ] = None,
    no_approx_jacobian_log: Annotated[
        bool,
        typer.Option(help="Do not approximate JL only at control point positions."),
    ] = False,
    landmarks_weight: Annotated[
        Optional[float],
        typer.Option(help="Weight for landmark distance regularisation (0-1)."),
    ] = None,
    landmarks_file: Annotated[
        Optional[Path],
        typer.Option(
            help="Landmark positions file (mm): refX refY refZ floX floY floZ per line."
        ),
    ] = None,
    # Similarity measures
    use_nmi: Annotated[
        bool, typer.Option(help="Force NMI even when other measures are specified.")
    ] = False,
    reference_bins: Annotated[
        Optional[int], typer.Option(help="NMI: number of bins for reference histogram.")
    ] = None,
    floating_bins: Annotated[
        Optional[int], typer.Option(help="NMI: number of bins for floating histogram.")
    ] = None,
    lncc_sigma: Annotated[
        Optional[float],
        typer.Option(help="Use LNCC with this Gaussian kernel std dev."),
    ] = None,
    use_ssd: Annotated[
        bool, typer.Option(help="Use SSD (images normalised to 0-1).")
    ] = False,
    use_ssd_no_norm: Annotated[
        bool, typer.Option(help="Use SSD without normalisation.")
    ] = False,
    mind_offset: Annotated[
        Optional[int], typer.Option(help="Use MIND descriptor with this offset.")
    ] = None,
    mindssc_offset: Annotated[
        Optional[int], typer.Option(help="Use MIND-SSC descriptor with this offset.")
    ] = None,
    use_kld: Annotated[
        bool, typer.Option(help="Use Kullback-Leibler divergence.")
    ] = False,
    similarity_weight_image: Annotated[
        Optional[Path], typer.Option(help="Per-voxel weight image for similarity.")
    ] = None,
    robust_range: Annotated[
        bool,
        typer.Option(help="Threshold intensities between 2nd and 98th percentile."),
    ] = False,
    # Optimisation
    max_iterations: Annotated[
        Optional[int], typer.Option(help="Max iterations at the final level. [150]")
    ] = None,
    num_levels: Annotated[
        Optional[int], typer.Option(help="Number of levels for pyramids. [3]")
    ] = None,
    num_levels_to_perform: Annotated[
        Optional[int], typer.Option(help="Number of levels to run. [ln]")
    ] = None,
    no_pyramid: Annotated[
        bool, typer.Option(help="Do not use a pyramidal approach.")
    ] = False,
    no_conjugate_gradient: Annotated[
        bool,
        typer.Option(help="Use simple gradient ascent instead of conjugate gradient."),
    ] = False,
    perturbation_steps: Annotated[
        Optional[int],
        typer.Option(help="Add perturbation step(s) after each optimisation."),
    ] = None,
    # F3D2 options
    velocity_field: Annotated[
        bool, typer.Option(help="Use velocity field integration for deformation.")
    ] = False,
    no_gradient_accumulation: Annotated[
        bool, typer.Option(help="Disable gradient accumulation through exponentiation.")
    ] = False,
    floating_mask: Annotated[
        Optional[Path], typer.Option(help="Mask image in the floating space.")
    ] = None,
    # Other
    smooth_gradient: Annotated[
        Optional[float], typer.Option(help="Smooth the metric derivative (in mm). [0]")
    ] = None,
    padding: Annotated[
        Optional[float], typer.Option(help="Padding value. [nan]")
    ] = None,
    verbose_off: Annotated[bool, typer.Option(help="Turn verbose off.")] = False,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of threads to use with OpenMP.")
    ] = None,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            is_eager=True,
            callback=_version_callback,
            help="Print version and exit.",
        ),
    ] = False,
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_f3d help and exit.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Fast Free-Form Deformation (F3D) non-rigid registration."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_f3d")

    if (landmarks_weight is not None) != (landmarks_file is not None):
        tool_logger.error(
            "Both --landmarks-weight and --landmarks-file must be provided together."
        )
        raise typer.Exit(code=1)

    # Check if the default output file exists before registration
    default_output = Path("outputCPP.nii")
    existed_before = default_output.exists()

    # Determine if user requested the default output path
    user_requested_default = False
    if output_cpp is not None:
        try:
            # Resolve both paths to absolute paths for comparison
            user_path = Path(output_cpp).resolve()
            default_path = default_output.resolve()
            user_requested_default = user_path == default_path
        except (OSError, RuntimeError):
            # If path resolution fails, treat as different paths
            pass

    args: list[str] = ["-ref", str(reference), "-flo", str(floating)]
    # Initial transformation
    if input_affine is not None:
        args.extend(["-aff", str(input_affine)])
    if input_cpp is not None:
        args.extend(["-incpp", str(input_cpp)])
    # Output
    if output_cpp is not None:
        args.extend(["-cpp", str(output_cpp)])
    if output_result is not None:
        args.extend(["-res", str(output_result)])
    # Input image options
    if reference_mask is not None:
        args.extend(["-rmask", str(reference_mask)])
    if smooth_reference is not None:
        args.extend(["-smooR", str(smooth_reference)])
    if smooth_floating is not None:
        args.extend(["-smooF", str(smooth_floating)])
    if reference_lower_threshold is not None:
        args.extend(["--rLwTh", str(reference_lower_threshold)])
    if reference_upper_threshold is not None:
        args.extend(["--rUpTh", str(reference_upper_threshold)])
    if floating_lower_threshold is not None:
        args.extend(["--fLwTh", str(floating_lower_threshold)])
    if floating_upper_threshold is not None:
        args.extend(["--fUpTh", str(floating_upper_threshold)])
    # Spline
    if spacing_x is not None:
        args.extend(["-sx", str(spacing_x)])
    if spacing_y is not None:
        args.extend(["-sy", str(spacing_y)])
    if spacing_z is not None:
        args.extend(["-sz", str(spacing_z)])
    # Regularisation
    if bending_energy is not None:
        args.extend(["-be", str(bending_energy)])
    if linear_energy is not None:
        args.extend(["-le", str(linear_energy)])
    if jacobian_log_weight is not None:
        args.extend(["-jl", str(jacobian_log_weight)])
    if no_approx_jacobian_log:
        args.append("-noAppJL")
    if landmarks_weight is not None and landmarks_file is not None:
        args.extend(["-land", str(landmarks_weight), str(landmarks_file)])
    # Similarity
    if use_nmi:
        args.append("--nmi")
    if reference_bins is not None:
        args.extend(["--rbn", str(reference_bins)])
    if floating_bins is not None:
        args.extend(["--fbn", str(floating_bins)])
    if lncc_sigma is not None:
        args.extend(["--lncc", str(lncc_sigma)])
    if use_ssd:
        args.append("--ssd")
    if use_ssd_no_norm:
        args.append("--ssdn")
    if mind_offset is not None:
        args.extend(["--mind", str(mind_offset)])
    if mindssc_offset is not None:
        args.extend(["--mindssc", str(mindssc_offset)])
    if use_kld:
        args.append("--kld")
    if similarity_weight_image is not None:
        args.extend(["-wSim", str(similarity_weight_image)])
    if robust_range:
        args.append("-rr")
    # Optimisation
    if max_iterations is not None:
        args.extend(["-maxit", str(max_iterations)])
    if num_levels is not None:
        args.extend(["-ln", str(num_levels)])
    if num_levels_to_perform is not None:
        args.extend(["-lp", str(num_levels_to_perform)])
    if no_pyramid:
        args.append("-nopy")
    if no_conjugate_gradient:
        args.append("-noConj")
    if perturbation_steps is not None:
        args.extend(["-pert", str(perturbation_steps)])
    # F3D2
    if velocity_field:
        args.append("-vel")
    if no_gradient_accumulation:
        args.append("-nogce")
    if floating_mask is not None:
        args.extend(["-fmask", str(floating_mask)])
    # Other
    if smooth_gradient is not None:
        args.extend(["-smoothGrad", str(smooth_gradient)])
    if padding is not None:
        args.extend(["-pad", str(padding)])
    if verbose_off:
        args.append("-voff")
    if omp_threads is not None:
        args.extend(["-omp", str(omp_threads)])

    run("reg_f3d", *args, tool_logger=tool_logger)

    # Clean up the default output file if it was created and not requested
    if not existed_before and default_output.exists() and not user_requested_default:
        try:
            default_output.unlink()
            logger.bind(executable="niftyregw").debug(
                f"Cleaned up default output file: {default_output}"
            )
        except OSError as e:
            logger.bind(executable="niftyregw").warning(
                f"Failed to clean up {default_output}: {e}"
            )



"""CLI command for reg_resample."""

from pathlib import Path
from typing import Annotated, Optional

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, make_version_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

_help_callback = make_help_callback("reg_resample")
_version_callback = make_version_callback("reg_resample")


def resample(
    reference: Annotated[
        Path, typer.Option("--reference", "-r", help="Reference image filename.")
    ],
    floating: Annotated[
        Path, typer.Option("--floating", "-f", help="Floating image filename.")
    ],
    transformation: Annotated[
        Optional[Path],
        typer.Option(
            "--transformation",
            "-t",
            help="Transformation file (from reg_aladin, reg_f3d, or reg_transform).",
        ),
    ] = None,
    output_result: Annotated[
        Optional[Path],
        typer.Option("--output-result", "-o", help="Resampled image filename."),
    ] = None,
    output_blank: Annotated[
        Optional[Path], typer.Option(help="Resampled blank grid filename.")
    ] = None,
    interpolation: Annotated[
        Optional[int],
        typer.Option(help="Interpolation order (0=NN, 1=LIN, 3=CUB, 4=SINC). [3]"),
    ] = None,
    padding: Annotated[
        Optional[float], typer.Option(help="Interpolation padding value. [0]")
    ] = None,
    tensor: Annotated[
        bool,
        typer.Option(
            help="Treat last 6 timepoints as tensor (XX, XY, YY, XZ, YZ, ZZ)."
        ),
    ] = False,
    psf: Annotated[
        bool, typer.Option(help="Two-step resampling for lower resolution.")
    ] = False,
    psf_algorithm: Annotated[
        Optional[int],
        typer.Option(
            help="PSF algorithm: minimise matrix metric (0) or determinant (1). [0]"
        ),
    ] = None,
    verbose_off: Annotated[bool, typer.Option(help="Turn verbose off.")] = False,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of threads to use with OpenMP.")
    ] = None,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            is_eager=True,
            callback=_version_callback,
            help="Print version and exit.",
        ),
    ] = False,
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_resample help and exit.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Resample an image with a given transformation."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_resample")

    args: list[str] = ["-ref", str(reference), "-flo", str(floating)]
    if transformation is not None:
        args.extend(["-trans", str(transformation)])
    if output_result is not None:
        args.extend(["-res", str(output_result)])
    if output_blank is not None:
        args.extend(["-blank", str(output_blank)])
    if interpolation is not None:
        args.extend(["-inter", str(interpolation)])
    if padding is not None:
        args.extend(["-pad", str(padding)])
    if tensor:
        args.append("-tensor")
    if psf:
        args.append("-psf")
    if psf_algorithm is not None:
        args.extend(["-psf_alg", str(psf_algorithm)])
    if verbose_off:
        args.append("-voff")
    if omp_threads is not None:
        args.extend(["-omp", str(omp_threads)])

    run("reg_resample", *args, tool_logger=tool_logger)



"""CLI command for reg_tools."""

from pathlib import Path
from typing import Annotated, Optional, Tuple

import typer
from loguru import logger

from niftyregw.commands import make_help_callback, make_version_callback, setup_logger
from niftyregw.enums import LogLevel
from niftyregw.wrapper import run

_help_callback = make_help_callback("reg_tools")
_version_callback = make_version_callback("reg_tools")


def tools(
    input_image: Annotated[
        Path, typer.Option("--input", "-i", help="Input image filename.")
    ],
    output: Annotated[
        Optional[Path],
        typer.Option("--output", "-o", help="Output image filename. [output.nii]"),
    ] = None,
    # Type conversion
    to_float: Annotated[
        bool, typer.Option(help="Convert the input image to float.")
    ] = False,
    # Geometric
    downsample: Annotated[
        bool, typer.Option(help="Downsample the input image 2 times.")
    ] = False,
    isotropic: Annotated[
        bool, typer.Option(help="Make the resulting image isotropic.")
    ] = False,
    change_resolution: Annotated[
        Optional[Tuple[float, float, float]],
        typer.Option(help="Resample to specified resolution in mm (x y z)."),
    ] = None,
    # Smoothing
    smooth_spline: Annotated[
        Optional[Tuple[float, float, float]],
        typer.Option(help="Smooth with cubic b-spline kernel (sx sy sz)."),
    ] = None,
    smooth_gaussian: Annotated[
        Optional[Tuple[float, float, float]],
        typer.Option(help="Smooth with Gaussian kernel (sx sy sz)."),
    ] = None,
    smooth_labels: Annotated[
        Optional[Tuple[float, float, float]],
        typer.Option(help="Smooth label image with Gaussian kernel (sx sy sz)."),
    ] = None,
    # Arithmetic
    add: Annotated[
        Optional[str], typer.Option(help="Add an image or value to the input.")
    ] = None,
    subtract: Annotated[
        Optional[str], typer.Option(help="Subtract an image or value from the input.")
    ] = None,
    multiply: Annotated[
        Optional[str], typer.Option(help="Multiply the input by an image or value.")
    ] = None,
    divide: Annotated[
        Optional[str], typer.Option(help="Divide the input by an image or value.")
    ] = None,
    # Thresholding and masking
    binarize: Annotated[
        bool, typer.Option(help="Binarise the input image (val!=0 -> 1).")
    ] = False,
    threshold: Annotated[
        Optional[float],
        typer.Option(help="Threshold the input image (val<thr -> 0, else 1)."),
    ] = None,
    nan_mask: Annotated[
        Optional[Path], typer.Option(help="Mask image; voxels outside are set to NaN.")
    ] = None,
    rms: Annotated[
        Optional[Path],
        typer.Option(help="Compute mean RMS between input and this image."),
    ] = None,
    remove_nan_inf: Annotated[
        Optional[float], typer.Option(help="Replace NaN and Inf with this value.")
    ] = None,
    # Header
    no_scaling: Annotated[
        bool, typer.Option(help="Set scl_slope=1 and scl_inter=0.")
    ] = False,
    # Format conversion
    to_rgb: Annotated[
        bool, typer.Option("--to-rgb", help="Convert 4D/5D to RGB nifti.")
    ] = False,
    # Descriptors
    mind: Annotated[bool, typer.Option(help="Create a MIND descriptor image.")] = False,
    mindssc: Annotated[
        bool, typer.Option(help="Create a MIND-SSC descriptor image.")
    ] = False,
    test_active_blocks: Annotated[
        bool,
        typer.Option(help="Highlight active blocks for reg_aladin (block variance)."),
    ] = False,
    # Other
    interpolation: Annotated[
        Optional[int], typer.Option(help="Interpolation order to warp the image.")
    ] = None,
    omp_threads: Annotated[
        Optional[int], typer.Option(help="Number of threads to use with OpenMP.")
    ] = None,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            is_eager=True,
            callback=_version_callback,
            help="Print version and exit.",
        ),
    ] = False,
    _: Annotated[
        bool,
        typer.Option(
            "--print-help",
            "-h",
            is_eager=True,
            callback=_help_callback,
            help="Print the original reg_tools help and exit.",
        ),
    ] = False,
    log_level: Annotated[
        LogLevel,
        typer.Option(
            "--log",
            case_sensitive=False,
            help="Set the log level.",
            rich_help_panel="Logging",
        ),
    ] = LogLevel.DEBUG,
) -> None:
    """Image manipulation tools."""
    setup_logger(log_level)
    tool_logger = logger.bind(executable="reg_tools")

    args: list[str] = ["-in", str(input_image)]
    if output is not None:
        args.extend(["-out", str(output)])
    if to_float:
        args.append("-float")
    if downsample:
        args.append("-down")
    if isotropic:
        args.append("-iso")
    if change_resolution is not None:
        args.extend(
            [
                "-chgres",
                str(change_resolution[0]),
                str(change_resolution[1]),
                str(change_resolution[2]),
            ]
        )
    if smooth_spline is not None:
        args.extend(
            [
                "-smoS",
                str(smooth_spline[0]),
                str(smooth_spline[1]),
                str(smooth_spline[2]),
            ]
        )
    if smooth_gaussian is not None:
        args.extend(
            [
                "-smoG",
                str(smooth_gaussian[0]),
                str(smooth_gaussian[1]),
                str(smooth_gaussian[2]),
            ]
        )
    if smooth_labels is not None:
        args.extend(
            [
                "-smoL",
                str(smooth_labels[0]),
                str(smooth_labels[1]),
                str(smooth_labels[2]),
            ]
        )
    if add is not None:
        args.extend(["-add", add])
    if subtract is not None:
        args.extend(["-sub", subtract])
    if multiply is not None:
        args.extend(["-mul", multiply])
    if divide is not None:
        args.extend(["-div", divide])
    if binarize:
        args.append("-bin")
    if threshold is not None:
        args.extend(["-thr", str(threshold)])
    if nan_mask is not None:
        args.extend(["-nan", str(nan_mask)])
    if rms is not None:
        args.extend(["-rms", str(rms)])
    if remove_nan_inf is not None:
        args.extend(["-rmNanInf", str(remove_nan_inf)])
    if no_scaling:
        args.append("-noscl")
    if to_rgb:
        args.append("-4d2rgb")
    if mind:
        args.append("-mind")
    if mindssc:
        args.append("-mindssc")
    if test_active_blocks:
        args.append("-testActiveBlocks")
    if interpolation is not None:
        args.extend(["-interp", str(interpolation)])
    if omp_threads is not None:
        args.extend(["-omp", str(omp_threads)])

    run("reg_tools", *args, tool_logger=tool_logger)



import platform
import shutil
import subprocess
import tempfile
import zipfile
from pathlib import Path

import requests
from loguru import logger

_GITHUB_URL = "https://github.com/KCL-BMEIS/niftyreg/releases/download/v2.0.0/NiftyReg-{name}-v2.0.0.zip"


def _is_cuda_available():
    try:
        result = subprocess.run(
            ["nvidia-smi"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return result.returncode == 0
    except FileNotFoundError:
        return False


def get_platform():
    """Get the detected platform name for NiftyReg binary selection.

    Returns:
        Platform name string, one of: "Ubuntu", "Ubuntu-CUDA", "macOS",
        "macOS-Intel", "Windows", or "Windows-CUDA".
    """
    system = platform.system()
    has_cuda = _is_cuda_available()
    match system:
        case "Linux":
            platform_name = "Ubuntu-CUDA" if has_cuda else "Ubuntu"
        case "Darwin":
            is_intel = platform.processor() == "i386" or platform.processor() == "i686"
            platform_name = "macOS-Intel" if is_intel else "macOS"
        case "Windows":
            platform_name = "Windows-CUDA" if has_cuda else "Windows"
        case _:
            raise Exception(f"Unsupported platform: {system}")
    return platform_name


def _get_download_url():
    platform_name = get_platform()
    return _GITHUB_URL.format(name=platform_name)


_DEFAULT_OUTPUT_DIR = Path.home() / ".local" / "bin"


def download_niftyreg(out_dir: Path = _DEFAULT_OUTPUT_DIR) -> list[Path]:
    """Download NiftyReg binaries and install them to *out_dir*.

    Args:
        out_dir: Directory where the binaries will be placed.
            Defaults to ``~/.local/bin``.

    Returns:
        List of paths to the installed binaries.
    """
    url = _get_download_url()
    logger.info(f"Downloading from {url}")
    response = requests.get(url)
    if response.status_code != 200:
        msg = f"Failed to download NiftyReg. Status code: {response.status_code}"
        raise RuntimeError(msg)

    zip_path = Path(tempfile.gettempdir(), "NiftyReg.zip")
    with open(zip_path, "wb") as f:
        f.write(response.content)
    out_tmp_dir = Path(tempfile.gettempdir(), "NiftyReg")
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(out_tmp_dir)

    out_dir.mkdir(parents=True, exist_ok=True)
    installed = []
    for path in out_tmp_dir.rglob("**/reg_*"):
        if not path.is_file():
            continue
        path.chmod(path.stat().st_mode | 0o111)
        dest = out_dir / path.name
        shutil.move(path, dest)
        installed.append(dest)

    # Clean up
    shutil.rmtree(out_tmp_dir, ignore_errors=True)
    zip_path.unlink(missing_ok=True)

    return sorted(installed)


def _which(program: str) -> Path | None:
    path = shutil.which(program)
    return Path(path) if path else None


BINARIES = (
    "reg_aladin",
    "reg_average",
    "reg_f3d",
    "reg_jacobian",
    "reg_measure",
    "reg_resample",
    "reg_tools",
    "reg_transform",
)


def find(tool: str) -> Path | None:
    """Find a NiftyReg binary by name (e.g. ``"reg_aladin"``)."""
    return _which(tool)


def aladin() -> Path | None:
    return find("reg_aladin")
